// avi2divx
// (C) 2000/10/23 Ulrich Hecht <uli@emulinks.de>
// last changed 2000/12/15
// licensed under the GNU Public License (GPL) v2

#define APPNAME "avi2divx"
#define APPVERSION "0.7"

#include <videoencoder.h>
#include <avifile.h>
#include <aviplay.h>
#include <aviutil.h>
#include <version.h>
#include <sys/time.h>

int main(int argc, char** argv)
{
	if(GetAvifileVersion()!=AVIFILE_VERSION)
	{
	        cout<<"This binary was compiled for Avifile ver. "<<AVIFILE_VERSION<<", but the library is ver. "<<GetAvifileVersion()<<". Aborting."<<endl;
	        return 0;
	}


	IAviReadFile* file;

	if(argc < 3)
	{
		printf("\nUsage: %s [OPTION]... [input AVI]... [output AVI]\n",argv[0]);
		printf("\nOptions:\n");
		printf("  -b\tDivX ;-) bitrate (default 1800)\n");
		printf("  -f\tuse fast-motion codec (default low-motion)\n");
		printf("  -a\tMP3 bitrate (kBit, default 112)\n");
		printf("  -e\tnumber of frames to encode (default all)\n");
		printf("  -c\tnumber of lines to clip from top/bottom (default none)\n");
		printf("  -o\toffset of clipping window (default 0 = center)\n");
		printf("  -bf\tradius for blur filter (default off)\n");
		printf("  -fa\taudio sample rate of input file (Hz); use only if avifile gets it wrong\n");
		printf("  -d\tdecimation (1:n scaling, default off)\n");
		//printf("  -g\tguess values for -c and -o options\n");
		printf("  -u\tflip picture (use if output video stream is upside-down, default off)\n");
		printf("\n" APPNAME " version " APPVERSION " by Ulrich Hecht <uli@emulinks.de>\n");
		exit(0);
	}

	int argbase=0;
	int divxbitrate=1800;
	int mp3bitrate=-1;
	int codec=fccDIV3;
	int endframe=-1;
	int cliplines=0;
	int clipoffset=0;
	int blur=0;
	int forcedaudiorate=0;
	int decimation=0;
	int guess=0;
	int flip=1;
	
	bool moreopts=true;

	while(moreopts)
	{
		moreopts=false;
		
		if(strcmp(argv[argbase+1],"-b")==0)
		{
			divxbitrate=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-f")==0)
		{
			codec=fccDIV4;
			argbase++;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-a")==0)
		{
			mp3bitrate=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}
		
		if(strcmp(argv[argbase+1],"-e")==0)
		{
			endframe=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-c")==0)
		{
			cliplines=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-o")==0)
		{
			clipoffset=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-bf")==0)
		{
			blur=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-fa")==0)
		{
			forcedaudiorate=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-d")==0)
		{
			decimation=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-g")==0)
		{
			guess=1;
			endframe=100;
			argbase++;
			moreopts=true;
		}
		if(strcmp(argv[argbase+1],"-u")==0)
		{
			flip=0;
			argbase++;
			moreopts=true;
		}
	}

	int numfiles=argc-argbase-2;
	char* inputfiles[numfiles];
	for(int i=argbase+1; i<argc-1; i++)
	{
		inputfiles[i-argbase-1]=argv[i];
		//printf("ifile %d assigned %s\n",i-argbase-1,argv[i]);
	}
	
	char* outputfile=argv[argc-1];
	
	file = CreateIAviReadFile(inputfiles[0]);
	
	IAviReadStream* invstream=file->GetStream(0, AviStream::Video);

        BITMAPINFOHEADER bh;
        invstream->StartStreaming();
        invstream->GetDecoder()->SetDirection(flip);
        invstream->GetDecoder()->SetDestFmt(24);
        invstream->GetOutputFormat(&bh, sizeof bh);
        bh.biWidth=abs(bh.biWidth);
        bh.biHeight=abs(bh.biHeight);

	if(decimation)
	{
		bh.biWidth/=decimation;
		bh.biHeight/=decimation;
	}

	bh.biHeight-=cliplines*2;
	int width=bh.biWidth;
	int height=bh.biHeight;

	unsigned char* framebuffer = new unsigned char[width*height*3];
	CImage imtarget((BitmapInfo*)&bh, framebuffer, false);

	double framerate=1/invstream->GetFrameTime();
	//printf("framerate: %d\n",framerate);
	int frames=invstream->GetEndPos();

	IAviReadStream* inastream=file->GetStream(0, AviStream::Audio);

	WAVEFORMATEX fmt;
	inastream->StartStreaming();
	inastream->GetOutputFormat(&fmt, sizeof fmt);
	
	int achan=fmt.nChannels;
	int rate=fmt.nSamplesPerSec;
	if(forcedaudiorate)
	{
		fmt.nSamplesPerSec=rate=forcedaudiorate;
	}

	printf("achan %d, rate %d\n",achan, rate);

	//unsigned char* framebuffer = new unsigned char[width*height*3];

	int sampsperframe = rate/framerate;
	//printf("--- sampsperframe: %d\n",sampsperframe);
	short* audiobuffer1 = new short[sampsperframe];
	short* audiobuffer2 = new short[sampsperframe];
	short* audiobufferint = new short[sampsperframe*2];

	int fccHandler=codec;

	IAviWriteFile* avifile;
	
	avifile=CreateIAviWriteFile(outputfile);

	IVideoEncoder::SetExtendedAttr(codec, "BitRate", divxbitrate);
	//IVideoEncoder::SetExtendedAttr(codec, "Crispness",99);
	
	IAviVideoWriteStream* stream=avifile->AddVideoStream(fccHandler, &bh, invstream->GetFrameTime()*1000000);
	stream->SetQuality(8000);
	
	stream->Start();
	
	if(mp3bitrate == -1) switch(rate) {
		case 48000: mp3bitrate=80*achan; break;
		case 44100: mp3bitrate=56*achan; break;
		case 22050: mp3bitrate=32*achan; break;
		default: printf("unknown sample rate %d Hz, please specify MP3 bitrate\n"); exit(1);
	}
	
	IAviAudioWriteStream* astream=avifile->AddAudioStream(0x55, &fmt, (mp3bitrate*1000)/8);
	astream->Start();

	Unsigned samples_read=0;
	Unsigned bytes_read=0;

	int totalframes=0;

	struct timeval tv;
	struct timezone tz={0,0};
	gettimeofday(&tv,&tz);
	long startsec=tv.tv_sec;
	long startusec=tv.tv_usec;
	double fps=0;

	int audiobytesperframe=invstream->GetFrameTime()*rate*achan*fmt.wBitsPerSample/8;
	printf("audiobytesperframe: %d\n",audiobytesperframe);
	int written_audio=0;

	CImage* oldimg=NULL;

	int lineguesstop=0;
	int lineguessbottom=0;
	int lineguesstopcount=0;
	int lineguessbottomcount=0;

	for (int j=0; j< numfiles; j++)
	{	
		if(j)
		{
			delete file;
			file = CreateIAviReadFile(inputfiles[j]);
			
			invstream=file->GetStream(0, AviStream::Video);
		
		        invstream->StartStreaming();
	        	invstream->GetDecoder()->SetDirection(flip);
        		invstream->GetDecoder()->SetDestFmt(24);
			
			frames=invstream->GetEndPos();
		
			inastream=file->GetStream(0, AviStream::Audio);
		
			inastream->StartStreaming();
		}
		
		for (int i=0; i< frames ; i++, totalframes++)
		{
			if(endframe > 0 && totalframes >= endframe) goto finished;
			
			//if(totalframes%10 == 0) printf("Encoding frame %d...\n",totalframes);
			gettimeofday(&tv,&tz);
			fps=totalframes/((tv.tv_sec+tv.tv_usec/1000000.0)-(startsec+startusec/1000000.0));
			
			if(!guess) { printf("Encoding frame %d, %.1lf frames per sec    \r",totalframes,fps); fflush(stdout); }
			invstream->ReadFrame();
			CImage* imsrc=invstream->GetFrame();
	
			if(blur)
			{
				//imsrc->ToYUV();
				imsrc->Blur(blur);
				//imsrc->ToRGB();
			}

			if(imsrc->yuv()) imsrc->ToRGB();

			if(!decimation)
				memcpy(framebuffer,imsrc->data()+(cliplines-clipoffset)*width*3,width*height*3);
			else
			{
				for(int y=0;y<height;y++)
					for(int x=0;x<width;x++)
						memcpy(framebuffer+x*3+y*width*3,imsrc->data()+x*decimation*3+(y+cliplines-clipoffset)*decimation*width*decimation*3,3);
			}

			if(guess)
			{
				double totallum;
				double oldlum=0;
				for(int y=0;y<height;y++)
				{
					totallum=0;
					for(int x=0;x<width*3;x++)
					{
						totallum+=*(framebuffer+y*width*3+x);
					}
					totallum /= 3*width;
					//printf("line %d: totallum: %lf, oldlum: %lf\n",y,totallum,oldlum);
					if(totallum > oldlum + 30)
					{
						printf("top line: %d\n",y);
						lineguesstop+=y;
						lineguesstopcount++;
						break;
					}
					oldlum=totallum;
				}
				for(int y=height;y>0;y--)
				{
					totallum=0;
					for(int x=0;x<width*3;x++)
					{
						totallum+=*(framebuffer+y*width*3+x);
					}
					totallum /= 3*width;
					if(totallum > oldlum + 30)
					{
						printf("bottom line: %d\n",y);
						lineguessbottom+=y;
						lineguessbottomcount++;
						break;
					}
					oldlum=totallum;
				}
				continue;
			}
			
 			stream->AddFrame(&imtarget);
			
			long long excess=(long long)(audiobytesperframe*(totalframes+1)-written_audio);
			while((excess>0) && (!inastream->Eof()))
			{
				inastream->ReadFrames(audiobufferint, (excess>sampsperframe*achan*2)?excess:sampsperframe*achan*2, sampsperframe*achan*2, samples_read, bytes_read);
				written_audio+=bytes_read;
				excess-=bytes_read;
				//printf("samples_read %d, bytes_read %d\n",samples_read,bytes_read);
				astream->AddData(audiobufferint,bytes_read);
			}
	
			//printf("framebuffer = %d %d %d\n",framebuffer[0],framebuffer[1],framebuffer[2]);
 			//printf("Writing frame %d...\n",i);
	
		}
	}
finished:
	stream->Stop();
	delete avifile;
	delete file;

	if (guess)
	{
		int avgtop=lineguesstop/lineguesstopcount;
		int avgbottom=lineguessbottom/lineguessbottomcount;
		printf("avg line top: %lf\n",(double)lineguesstop/lineguesstopcount);
		printf("avg line bottom: %lf\n",(double)lineguessbottom/lineguessbottomcount);
		printf("suggested options: -c %d -o %d\n",(avgtop+height-avgbottom)/2,-avgtop+((avgtop+height-avgbottom)/2));
	}
}
