// mpeg2divx
// 2000/10/23 Ulrich Hecht <uli@emulinks.de>
// last changed 2000/12/15
// licensed under the GNU Public License (GPL) v2

#define APPNAME "mpeg2divx"
#define APPVERSION "0.7"

#include <videoencoder.h>
#include <avifile.h>
#include <aviplay.h>
#include <aviutil.h>
#include <version.h>
#include <libmpeg3.h>
#include <sys/time.h>

int main(int argc, char** argv)
{
	if(GetAvifileVersion()!=AVIFILE_VERSION)
	{
	        cout<<"This binary was compiled for Avifile ver. "<<AVIFILE_VERSION<<", but the library is ver. "<<GetAvifileVersion()<<". Aborting."<<endl;
	        return 0;
	}

	mpeg3_t* file;

	if(argc < 3)
	{
		printf("\nUsage: %s [OPTION]... [input MPEG]... [output AVI]\n",argv[0]);
		printf("\nOptions:\n");
		printf("  -b\tDivX ;-) bitrate (default 1800)\n");
		printf("  -f\tuse fast-motion codec (default low-motion)\n");
		printf("  -a\tMP3 bitrate (kBit, default 112)\n");
                printf("  -e\tnumber of frames to encode (default all)\n");
                printf("  -c\tnumber of lines to clip from top/bottom (default none)\n");
		printf("  -o\toffset of clipping window (default 0 = center)\n");
                printf("  -bf\tradius for blur filter (default off)\n");
		printf("  -d\tdecimation (1:n scaling, default off)\n");
		printf("  -l\tsource audio channel for left target channel (default 0)\n");
		printf("  -r\tsource audio channel for right target channel (default 1)\n");
		printf("\n" APPNAME " version " APPVERSION " by Ulrich Hecht <uli@emulinks.de>\n");
		exit(0);
	}

	int argbase=0;
	int divxbitrate=1800;
	int mp3bitrate=-1;
	int codec=fccDIV3;
	int endframe=-1;
	int cliplines=0;
	int clipoffset=0;
	int blur=0;
	int decimation=0;
	int leftaudio=0;
	int rightaudio=1;

	bool moreopts=true;

	while(moreopts)
	{
		moreopts=false;
		
		if(strcmp(argv[argbase+1],"-b")==0)
		{
			divxbitrate=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-f")==0)
		{
			codec=fccDIV4;
			argbase++;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-a")==0)
		{
			mp3bitrate=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}
		
		if(strcmp(argv[argbase+1],"-e")==0)
		{
			endframe=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-c")==0)
		{
			cliplines=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-o")==0)
		{
			clipoffset=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-bf")==0)
		{
			blur=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-d")==0)
		{
			decimation=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-l")==0)
		{
			leftaudio=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}

		if(strcmp(argv[argbase+1],"-r")==0)
		{
			rightaudio=atoi(argv[argbase+2]);
			argbase+=2;
			moreopts=true;
		}
	}

	int numfiles=argc-argbase-2;
	char* inputfiles[numfiles];
	for(int i=argbase+1; i<argc-1; i++)
	{
		inputfiles[i-argbase-1]=argv[i];
		printf("ifiles %d assigned %s\n",i-argbase-1,argv[i]);
	}
	
	char* outputfile=argv[argc-1];
	
	file = mpeg3_open(inputfiles[0]);
	
	int achan=mpeg3_audio_channels(file, 0);
	int rate=mpeg3_sample_rate(file, 0);
	long samples=mpeg3_audio_samples(file, 0);
	printf("channels %d, rate %d, samples %d\n",achan,rate,samples);
	
	int width=mpeg3_video_width(file, 0);
	int height=mpeg3_video_height(file, 0);
	float framerate=mpeg3_frame_rate(file, 0);
	int frames=mpeg3_video_frames(file, 0);
	//printf("width %d, height %d, framerate %f, frames %d\n",width,height,framerate,frames);

	int scaledheight,scaledwidth;
	unsigned char* scaledbuffer;
	if(decimation)
	{
		scaledheight=height/decimation-cliplines*2;
		scaledwidth=width/decimation;
		scaledbuffer= new unsigned char[scaledwidth*scaledheight*3];
	}

	unsigned char* framebuffer = new unsigned char[width*height*3];
	unsigned char* extrabuffer = new unsigned char[width*3+4];
	unsigned char** rowptr = new (unsigned char*)[height];
	for (int i=0; i<height; i++)
	{
		if(i==height-1)
		{
			rowptr[i]=extrabuffer;
		}
		rowptr[i]=framebuffer+(width*(height-1)*3)-(width*3*i);
	}

	height-=cliplines*2;
	//printf("framebuffer: %d",(unsigned int)framebuffer);

	int sampsperframe = rate/framerate;
	short* audiobuffer1 = new short[sampsperframe];
	short* audiobuffer2 = new short[sampsperframe];
	short* audiobufferint = new short[sampsperframe*2];
	int fccHandler=codec;

	IAviWriteFile* avifile;
	
	avifile=CreateIAviWriteFile(outputfile);

	IVideoEncoder::SetExtendedAttr(codec, "BitRate", divxbitrate);
	
	BITMAPINFOHEADER bh;
	memset(&bh,0,sizeof(BITMAPINFOHEADER));
	bh.biSize=sizeof(BITMAPINFOHEADER);
	bh.biWidth=width;
	bh.biHeight=height;
	bh.biPlanes=1;
	bh.biBitCount=24;
	bh.biCompression=0;
	bh.biSizeImage=width*height*3;

	CImage* im;
	if(!decimation)
		im=new CImage((BitmapInfo*)&bh, framebuffer+(cliplines-clipoffset)*width*3, false);
	else
	{
		bh.biWidth=scaledwidth;
		bh.biHeight=scaledheight;
		bh.biSizeImage=scaledwidth*scaledheight*3;
		im=new CImage((BitmapInfo*)&bh, scaledbuffer, false);
	}

	IAviVideoWriteStream* stream=avifile->AddVideoStream(fccHandler, &bh, 1000000/framerate);
	stream->SetQuality(8000);
	
	stream->Start();
	
	WAVEFORMATEX fmt;
	memset(&fmt,0,sizeof(WAVEFORMATEX));
	fmt.wFormatTag=1;
	fmt.nChannels=(achan>2 ? 2 : achan);
	fmt.nSamplesPerSec=rate;
	fmt.nAvgBytesPerSec=(achan>1 ? rate : rate*2);
	fmt.nBlockAlign=(achan>1 ? 2 : 1);
	fmt.wBitsPerSample=16;
	fmt.cbSize=sizeof(WAVEFORMATEX);

	if(achan>0 && mp3bitrate == -1) switch(rate) {
		case 48000: mp3bitrate=(achan>1?160:80*achan); break;
		case 44100: mp3bitrate=(achan>1?112:56*achan); break;
		case 22050: mp3bitrate=(achan>1?64:32*achan); break;
		default: printf("unknown sample rate %d Hz, please specify MP3 bitrate\n"); exit(1);
	}
	
	IAviAudioWriteStream* astream;
	if(achan>0)
	{
		astream=avifile->AddAudioStream(0x55, &fmt, (mp3bitrate*1000)/8);
		astream->Start();
	}

	int samples_read=0;

	int totalframes=0;

	struct timeval tv;
	struct timezone tz={0,0};
	gettimeofday(&tv,&tz);
	long startsec=tv.tv_sec;
	long startusec=tv.tv_usec;
	double fps=0;
	for (int j=0; j< numfiles; j++)
	{	
		if(j)
		{
			mpeg3_close(file);
			printf("opening MPG %s\n",inputfiles[j]);
			file = mpeg3_open(inputfiles[j]);
			frames=mpeg3_video_frames(file, 0);
		}
		for (int i=0; i< frames-15 ; i++, totalframes++)
		{
			if(endframe > 0 && totalframes >= endframe) goto finished;

			//if(totalframes%10 == 0) printf("Encoding frame %d...\n",totalframes);
			gettimeofday(&tv,&tz);
			fps=totalframes/((tv.tv_sec+tv.tv_usec/1000000.0)-(startsec+startusec/1000000.0));
			printf("Encoding frame %d, %.1lf frames per sec    \r",totalframes,fps); fflush(stdout);

	 		if(mpeg3_read_frame(file, rowptr, 0, 0, width, height+cliplines-clipoffset, width, height+cliplines-clipoffset, MPEG3_BGR888, 0))
			{
				printf("error reading frame %d\n",i);
				exit(1);
			}
			memcpy(framebuffer, extrabuffer, width*3);

			if(achan>0)
			{
				if(mpeg3_read_audio(file, NULL, audiobuffer1, leftaudio, sampsperframe, 0))
				{
					printf("error reading left audio (channel %d)\n",leftaudio);
					exit(2);
				}
				if(achan>1)
				{
					if(mpeg3_reread_audio(file, NULL, audiobuffer2, rightaudio, sampsperframe, 0))
					{
						printf("error reading right audio (channel %d)\n",rightaudio);
					exit(3);
					}
				}
				//printf("audiobuffer = %d %d %d\n",audiobuffer1[0], audiobuffer1[1], audiobuffer1[2]);
				//memset(framebuffer,i%256,352*288*3);
				if(achan>1) for(int j=0; j<sampsperframe; j++)
				{
					audiobufferint[j*2]=audiobuffer1[j];
					audiobufferint[j*2+1]=audiobuffer2[j];
				}
	
				if(achan>1)
					astream->AddData(audiobufferint,sampsperframe*2);
				else
					astream->AddData(audiobuffer1,sampsperframe);
			}

			//printf("framebuffer = %d %d %d\n",framebuffer[0],framebuffer[1],framebuffer[2]);
 			//printf("Writing frame %d...\n",i);
			if(decimation)
			{
				for(int y=0;y<scaledheight;y++)
					for(int x=0;x<scaledwidth;x++)
						memcpy(scaledbuffer+x*3+y*scaledwidth*3,framebuffer+x*decimation*3+(y+cliplines-clipoffset)*decimation*width*3,3);
			}
			if(blur)
			{
				im->ToYUV();
				im->Blur(blur);
			}
 			stream->AddFrame(im);

			samples_read+=sampsperframe;
			//printf("samples: %d, samples_read: %d\n",samples,samples_read);
		}
	}
finished:
	stream->Stop();
	delete avifile;
	mpeg3_close(file);
}
